//
// Copyright 2000, Kirk Davis (kbd@ndx.net)
//
// Use and distribute freely. Share your changes for others.
// Use at your own risk - no warrantee expressed or implied
// (but you knew that).
//


#include "1802.h"

#include "stdio.h"
#include "conio.h"
#include "ctype.h"

struct Regs Regs;

unsigned char Mem[0x10000];

char *Instructions[] = {
	"IDLE", "LDN 1", "LDN 2", "LDN 3", "LDN 4", "LDN 5", "LDN 6", "LDN 7",
	"LDN 8", "LDN 9", "LDN A", "LDN B", "LDN C", "LDN D", "LDN E", "LDN F",
	"INC R0", "INC R1", "INC R2", "INC R3", "INC R4", "INC R5", "INC R6", "INC R7",
	"INC R8", "INC R9", "INC RA", "INC RB", "INC RC", "INC RD", "INC RE", "INC RF",
	"DEC R0", "DEC R1", "DEC R2", "DEC R3", "DEC R4", "DEC R5", "DEC R6", "DEC R7",
	"DEC R8", "DEC R9", "DEC RA", "DEC RB", "DEC RC", "DEC RD", "DEC RE", "DEC RF",
	"BR", "BQ", "BZ", "BDF", "B1", "B2", "B3", "B4",
	"SKP", "BNQ", "BNZ", "BNF", "BN1", "BN2", "BN3", "BN4",
	"LDA R0", "LDA R1", "LDA R2", "LDA R3", "LDA R4", "LDA R5", "LDA R6", "LDA R7",
	"LDA R8", "LDA R9", "LDA RA", "LDA RB", "LDA RC", "LDA RD", "LDA RE", "LDA RF",
	"STR R0", "STR R1", "STR R2", "STR R3", "STR R4", "STR R5", "STR R6", "STR R7",
	"STR R8", "STR R9", "STR RA", "STR RB", "STR RC", "STR RD", "STR RE", "STR RF",
	"IRX", "OUT 1", "OUT 2", "OUT 3", "OUT 4", "OUT 5", "OUT 6", "OUT 7",
	"INVALID", "INP 1", "INP 2", "INP 3", "INP 4", "INP 5", "INP 6", "INP 7",
	"RET", "DIS", "LDXA", "STXD", "ADC", "SDC", "SHRC", "SMB",
	"SAV", "MARK", "REQ", "SEQ", "ADCI", "SDBI", "SHLC", "SMBI",
	"GLO R0", "GLO R1", "GLO R2", "GLO R3", "GLO R4", "GLO R5", "GLO R6", "GLO R7",
	"GLO R8", "GLO R9", "GLO RA", "GLO RB", "GLO RC", "GLO RD", "GLO RE", "GLO RF",
	"GHI R0", "GHI R1", "GHI R2", "GHI R3", "GHI R4", "GHI R5", "GHI R6", "GHI R7",
	"GHI R8", "GHI R9", "GHI RA", "GHI RB", "GHI RC", "GHI RD", "GHI RE", "GHI RF",
	"PLO R0", "PLO R1", "PLO R2", "PLO R3", "PLO R4", "PLO R5", "PLO R6", "PLO R7",
	"PLO R8", "PLO R9", "PLO RA", "PLO RB", "PLO RC", "PLO RD", "PLO RE", "PLO RF",
	"PHI R0", "PHI R1", "PHI R2", "PHI R3", "PHI R4", "PHI R5", "PHI R6", "PHI R7",
	"PHI R8", "PHI R9", "PHI RA", "PHI RB", "PHI RC", "PHI RD", "PHI RE", "PHI RF",
	"LBR", "LBQ", "LBZ", "LBDF", "NOP", "LSNQ", "LSNZ", "LSNF",
	"NLBR", "LBNQ", "LBNZ", "LBNF", "LSIE", "LSQ", "LSZ", "LSDF",
	"SEP 0", "SEP 1", "SEP 2", "SEP 3", "Call", "Ret", "SEP 6", "SEP 7",
	"SEP 8", "SEP 9", "SEP A", "SEP B", "SEP C", "SEP D", "SEP E", "SEP F",
	"SEX 0", "SEX 1", "SEX 2", "SEX 3", "SEX 4", "SEX 5", "SEX 6", "SEX 7",
	"SEX 8", "SEX 9", "SEX A", "SEX B", "SEX C", "SEX D", "SEX E", "SEX F",
	"LDX", "OR", "AND", "XOR", "ADD", "SD", "SHR", "SM",
	"LDI", "ORI", "ANI", "XRI", "ADI", "SDI", "SHL", "SMI"
};

unsigned char InstructionLength[] = {
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x00
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x08
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x10
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x18
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x20
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x28
	2, 2, 2, 2, 2, 2, 2, 2,		// 0x30
	1, 2, 2, 2, 2, 2, 2, 2,		// 0x38
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x40
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x48
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x50
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x58
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x60
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x68
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x70
	1, 1, 1, 1, 2, 2, 1, 2,		// 0x78
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x80
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x88
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x90
	1, 1, 1, 1, 1, 1, 1, 1,		// 0x98
	1, 1, 1, 1, 1, 1, 1, 1,		// 0xA0
	1, 1, 1, 1, 1, 1, 1, 1,		// 0xA8
	1, 1, 1, 1, 1, 1, 1, 1,		// 0xB0
	1, 1, 1, 1, 1, 1, 1, 1,		// 0xB8
	3, 3, 3, 3, 1, 1, 1, 1,		// 0xC0
	1, 3, 3, 3, 1, 1, 1, 1,		// 0xC8
	1, 1, 1, 1, 3, 1, 1, 1,		// 0xD0
	1, 1, 1, 1, 1, 1, 1, 1,		// 0xD8
	1, 1, 1, 1, 1, 1, 1, 1,		// 0xE0
	1, 1, 1, 1, 1, 1, 1, 1,		// 0xE8
	1, 1, 1, 1, 1, 1, 1, 1,		// 0xF0
	2, 2, 2, 2, 2, 2, 1, 2		// 0xF8
};

unsigned char DisAsm(unsigned char *buf, unsigned short offset) {
	if (InstructionLength[M(offset)] == 1) {
		sprintf(buf, "%0.4x %s\n", offset, Instructions[M(offset)]);
	} else if (InstructionLength[M(offset)] == 2) {
		sprintf(buf, "%0.4x %s %0.2x\n", offset, Instructions[M(offset)], M(offset+1));
	} else {
		sprintf(buf, "%0.4x %s %0.2x%0.2x\n", offset, Instructions[M(offset)], M(offset+1), M(offset+2));
	}
	return InstructionLength[M(offset)];
}


Step() {
	char iBuf[80];
	unsigned long TempLong;
	unsigned char RVal = 0;

#if DEBUG
	unsigned char d = D;
	unsigned char p = P;
	unsigned char df = DF;
	unsigned char x = X;
	unsigned short r0 = R(0);
	unsigned short r1 = R(1);
	unsigned short r2 = R(2);
	unsigned short r3 = R(3);
	unsigned short r4 = R(4);
	unsigned short r5 = R(5);
	unsigned short r6 = R(6);
	unsigned short r7 = R(7);
	unsigned short r8 = R(8);
	unsigned short r9 = R(9);
	unsigned short ra = R(0xa);
	unsigned short rb = R(0xb);
	unsigned short rc = R(0xc);
	unsigned short rd = R(0xd);
	unsigned short re = R(0xe);
	unsigned short rf = R(0xf);
	unsigned short *dfp = &DF;
#endif

	N = (Mem[PC] & 0xf);

	switch (Mem[PC++]) {
	case 0x00: // IDLE - WAIT FOR DMA OR INTERRUPT; M(R(0)) -> BUS
		break;

	case 0x01: // LDN	M(R(N) -> D
	case 0x02:
	case 0x03:
	case 0x04:
	case 0x05:
	case 0x06:
	case 0x07:
	case 0x08:
	case 0x09:
	case 0x0a:
	case 0x0b:
	case 0x0c:
	case 0x0d:
	case 0x0e:
	case 0x0f:
		D = M(R(N));
		break;

	case 0x10: // INC	R(N) + 1 -> R(N)
	case 0x11: 
	case 0x12:
	case 0x13:
	case 0x14:
	case 0x15:
	case 0x16:
	case 0x17:
	case 0x18:
	case 0x19:
	case 0x1a:
	case 0x1b:
	case 0x1c:
	case 0x1d:
	case 0x1e:
	case 0x1f:
		R(N)++;
		break;

	case 0x20: // DEC	R(N) - 1 -> R(N)
	case 0x21: 
	case 0x22:
	case 0x23:
	case 0x24:
	case 0x25:
	case 0x26:
	case 0x27:
	case 0x28:
	case 0x29:
	case 0x2a:
	case 0x2b:
	case 0x2c:
	case 0x2d:
	case 0x2e:
	case 0x2f:
		R(N)--;
		break;

	case 0x30: // BR	M(R(P)) -> R(P).0
		RL(P) = M(R(P));
		break;
	case 0x31: // BQ	IF Q = 1, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		if (Q == 1)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x32: // BZ	IF D = 0, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		if (D == 0)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x33: // BDF 33	IF DF = 1, M(R(P)) -> R(P).0, ELSE R(P) + ->(P)
		if (DF == 1)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x34: // B1 34		IF EF1 =1, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		if (EF1 == 1)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x3c: // BN1 3C	IF EF1 = 0, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		if (EF1 == 0)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x35: //B2 35		IF EF2 = 1, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		if (EF2 == 1)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x3d: //BN2 3D		IF EF2 = 0, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		if (EF2 == 0)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x36: //B3 36		IF EF3 = 1, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		if (EF3 == 1)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x3e: //BN3 3E		IF EF3 = 0, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		if (EF3 == 0)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x37: //B4 37		IF EF4 = 1, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		printf("EF4 ?", N);
		gets(iBuf);
		sscanf(iBuf, "%x", &EF4);

		if (EF4 == 1)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x3f: //BN4 3F		IF EF4 = 0, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		printf("EF4 ?");
		gets(iBuf);
		sscanf(iBuf, "%x", &EF4);

		if (EF4 == 0)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x38: //NBR 38		R(P) + 1 -> R(P)
			R(P)++;
		break;
	case 0x39: //BNQ 39		IF Q = 0, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		if (Q == 0)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x3a: //BNZ 3A		IF D NOT 0, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		if (D != 0)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;
	case 0x3b: //BNF 3B		IF DF = 0, M(R(P)) -> R(P).0, ELSE R(P) + 1 -> R(P)
		if (DF == 0)
			RL(P) = M(R(P));
		else
			R(P)++;
		break;

	case 0x40: // LDA 4N M(R(N)) -> D; R(N) + 1 -> R(N)
	case 0x41: 
	case 0x42:
	case 0x43:
	case 0x44:
	case 0x45:
	case 0x46:
	case 0x47:
	case 0x48:
	case 0x49:
	case 0x4a:
	case 0x4b:
	case 0x4c:
	case 0x4d:
	case 0x4e:
	case 0x4f:
		D = M(R(N));
		R(N)++;
		break;

	case 0x50: // STR 5N D -> M(R(N))
	case 0x51: 
	case 0x52:
	case 0x53:
	case 0x54:
	case 0x55:
	case 0x56:
	case 0x57:
	case 0x58:
	case 0x59:
	case 0x5a:
	case 0x5b:
	case 0x5c:
	case 0x5d:
	case 0x5e:
	case 0x5f:
		if (R(N) < 0x4000 || R(N) > 0x9000) 
			M(R(N)) = D;
		else
			RVal = 0;
		break;

	case 0x60: // IRX 60	R(X) + 1 -> R(X)
		break;

	case 0x68: // 68		Invalid
		if (M(R(P)) == 1) {				// PUTCHAR
			if (D == 13)
				putchar('\n');
			if (isprint(D))
				putchar(D);
		} else if (M(R(P)) == 2) {		// GETCHAR
			D = _getch();

			if (islower(D))
				D = toupper(D);

			if (D == 3) {
				RVal = 1;
				R(P)--;
				break;
			}
			
			putchar(D);
		}
		R(P)++;
		break;

	case 0x61: // OUT 1 61	M(R(X)) -> BUS; R(X) + 1 -> R(X); N LINES = 1
	case 0x62:
	case 0x63:
	case 0x64:
	case 0x65:
	case 0x66:
	case 0x67:
		printf("%0.2x\n", M(R(X)));
		D = M(R(X));
		R(X)++;
		break;
	
	case 0x69: // INP 1 69	BUS -> M(R(X)); BUS -> D; N LINES = 1
	case 0x6a:
	case 0x6b:
	case 0x6c:
	case 0x6d:
	case 0x6e:
	case 0x6f:
		printf("INP %d ?", N & 7);
		gets(iBuf);
		sscanf(iBuf, "%x", &TempLong);
		M(R(X)) = (unsigned char) TempLong;
		D = (unsigned char) TempLong;
		break;

	
	case 0x70: // RET 70	M(R(X)) -> (X, P); R(X) + 1 -> R(X), 1 -> lE
		return 1; // DEBUG
		X = M(R(X)) >> 4;
		P = M(R(X)) & 0xf;
		R(X)++;
		IE = 1;
		RVal = 1;
		break;
	case 0x71: // DlS 71	M(R(X)) -> (X, P); R(X) + 1 -> R(X), 0 -> lE
		X = M(R(X)) >> 4;
		P = M(R(X)) & 0xf;
		R(X)++;
		IE = 0;
		RVal = 1;
		break;
	case 0x78: // SAV 78	T -> M(R(X))
		P = T >> 4;
		X = T & 0xf;
		break;
	case 0x79: // MARK 79	(X, P) -> T; (X, P) -> M(R(2)), THEN P -> X; R(2) - 1 -> R(2)
		T = (X << 4) & P;
		if (R(2) < 0x4000 || R(2) > 0x9000)
			M(R(2)) = T;
		X = P;
		R(2)--;
		RVal = 1;
		break;
	case 0x7a: // REQ 7A	0 -> Q
		printf("Q0\n");
		Q = 0;
		break;
	case 0x7b: // SEQ 7B	1 -> Q
		printf("Q1\n");
		Q = 1;
		break;
	case 0x7f: // SMBl 7F	D-M(R(P))-(NOT DF) -> DF, D; R(P) + 1 -> R(P)
		TempLong = D - M(R(P)) - ((~DF)&1);
		DF = D - ((~DF)&1) < M(R(P))  ? 0 : 1;
		D = (unsigned char) TempLong;
		R(P)++;
		break;
	case 0x77: // SMB 77	D-M(R(X))-(NOT DF) -> DF, D
		TempLong = D - M(R(X)) - ((~DF)&1); //((~DF) & 1);
		DF = D - ((~DF)&1) < M(R(X))  ? 0 : 1;
		D = (unsigned char) TempLong;
		break;
	case 0x7d: // SDBl 7D	M(R(P)) - D - (Not DF) -> DF, D; R(P) + 1 -> R(P)
		TempLong =  M(R(P)) - D - ((~DF)&1);
		DF = M(R(P)) - ((~DF)&1) < D  ? 0 : 1;
		D = (unsigned char) TempLong;
		R(P)++;
		break;
	case 0x75: // SDB 75	M(R(X)) - D - (NOT DF) -> DF, D
		TempLong = M(R(X)) - D - ((~DF)&1);
		DF = M(R(X)) - ((~DF)&1) < D  ? 0 : 1;
		D = (unsigned char) TempLong;
		break;
	case 0x7c: // ADCl 7C	M(R(P)) + D + DF -> DF, D; R(P) + 1 -> R(P)
		TempLong =  M(R(P)) + D + DF;
		D = (unsigned char) TempLong;
		DF = TempLong >= 0x100 ? 1 : 0;
		R(P)++;
		break;
	case 0x74: // ADC 74	M(R(X)) + D + DF -> DF, D
		TempLong =  M(R(X)) + D + DF;
		D = (unsigned char) TempLong;
		DF = TempLong >= 0x100 ? 1 : 0;
		break;
	case 0x7e: // RSHL 7E	SHIFT D LEFT, MSB(D) -> DF, DF -> LSB(D)
		TempLong = (D << 1) | DF;
		D = (unsigned char) TempLong;
		DF = TempLong & 0x100 ? 1 : 0;
		break;
		
	case 0x76: // RSHR 76	SHIFT D RIGHT, LSB(D) -> DF, DF -> MSB(D)
		TempLong = (D >> 1) | (DF ? 0x80 : 0);
		DF = D & 1;
		D = (unsigned char) TempLong;
		break;

	case 0x73: // STXD 73	D -> M(R(X)); R(X) - 1 -> R(X)
		if (R(X) < 0x4000 || R(X) > 0x9000)
			M(R(X)) = D;
		else
			RVal = 0;
		R(X)--;
		break;
	case 0x72: // LDXA 72	M(R(X)) -> D; R(X) + 1 -> R(X)
		D = M(R(X));
		R(X)++;
		break;

	case 0x80: // GLO 8N	R(N).0 -> D
	case 0x81: 
	case 0x82:
	case 0x83:
	case 0x84:
	case 0x85:
	case 0x86:
	case 0x87:
	case 0x88:
	case 0x89:
	case 0x8a:
	case 0x8b:
	case 0x8c:
	case 0x8d:
	case 0x8e:
	case 0x8f:
		D = RL(N);
		break;

	case 0x90: // GHl 9N	R(N).1 -> D
	case 0x91: 
	case 0x92:
	case 0x93:
	case 0x94:
	case 0x95:
	case 0x96:
	case 0x97:
	case 0x98:
	case 0x99:
	case 0x9a:
	case 0x9b:
	case 0x9c:
	case 0x9d:
	case 0x9e:
	case 0x9f:
		D = RH(N);
		break;

	case 0xa0: // PLO AN	D -> R(N).0
	case 0xa1: 
	case 0xa2:
	case 0xa3:
	case 0xa4:
	case 0xa5:
	case 0xa6:
	case 0xa7:
	case 0xa8:
	case 0xa9:
	case 0xaa:
	case 0xab:
	case 0xac:
	case 0xad:
	case 0xae:
	case 0xaf:
		RL(N) = D;
		break;

	case 0xb0: // PHI BN	D -> R(N).1
	case 0xb1: 
	case 0xb2:
	case 0xb3:
	case 0xb4:
	case 0xb5:
	case 0xb6:
	case 0xb7:
	case 0xb8:
	case 0xb9:
	case 0xba:
	case 0xbb:
	case 0xbc:
	case 0xbd:
	case 0xbe:
	case 0xbf:
		RH(N) = D;
		break;

	case 0xc0: // LBR C0	M(R(P)) -> R(P). 1, M(R(P) + 1) -> R(P).0
		TempLong = M(R(P)++) << 8;
		TempLong |= M(R(P));
		R(P) = (unsigned short) TempLong;
		break;
	case 0xc1: // LBQ C1	IF Q = 1, M(R(P)) -> R(P).1, M(R(P) + 1) -> R(P).0,
		if (Q == 1) {
			TempLong = M(R(P)++) << 8;
			TempLong |= M(R(P));
			R(P) = (unsigned short) TempLong;
		} else {
			R(P) += 2;
		}
		break;
	case 0xc2: // LBZ C2	lF D = 0, M(R(P)) -> R(P).1, M(R(P) +1) -> R(P).0,
		if (D == 0) {
			TempLong = M(R(P)++) << 8;
			TempLong |= M(R(P));
			R(P) = (unsigned short) TempLong;
		} else {
			R(P) += 2;
		}
		break;
	case 0xc3: // LBDF C3	lF DF = 1, M(R(P)) -> R(P).1, M(R(P) + 1) -> R(P).0, ELSE
		if (DF == 1) {
			TempLong = M(R(P)++) << 8;
			TempLong |= M(R(P));
			R(P) = (unsigned short) TempLong;
		} else {
			R(P) += 2;
		}
		break;
	case 0xc4: // NOP C4	CONTINUE
		break;
	case 0xc5: // LSNQ C5	IF Q = 0, R(P) + 2 -> R(P), ELSE CONTINUE
		if (Q == 0)
			R(P) += 2;
		break;
	case 0xc6: // LSNZ C6	IF D Not 0, R(P) + 2 -> R(P), ELSE CONTINUE
		if (D != 0)
			R(P) += 2;
		break;
	case 0xc7: // LSNF C7	IF DF = 0, R(P) + 2 -> R(P), ELSE CONTINUE
		if (DF == 0)
			R(P) += 2;
		break;
	case 0xc8: // NLBR C8	R(P) = 2 -> R(P)
		R(P) += 2;
		break;
	case 0xc9: // LBNQ C9	lF Q = 0, M(R(P)) -> R(P).1, M(R(P) + 1) -> R(P).0
		if (Q == 0) {
			TempLong = M(R(P)++) << 8;
			TempLong |= M(R(P));
			R(P) = (unsigned short) TempLong;
		} else {
			R(P) += 2;
		}
		break;
	case 0xca: // LBNZ CA	IF D Not 0, M(R(P)) -> R(P).1, M(R(P) + 1) -> R(P).0, ELSE
		if (D != 0) {
			TempLong = M(R(P)++) << 8;
			TempLong |= M(R(P));
			R(P) = (unsigned short) TempLong;
		} else {
			R(P) += 2;
		}
		break;
	case 0xcb: // LBNF CB	IF DF = 0, M(R(P)) -> R(P).1, M(R(P) + 1) -> R(P).0, ELSE
		if (DF == 0) {
			TempLong = M(R(P)++) << 8;
			TempLong |= M(R(P));
			R(P) = (unsigned short) TempLong;
		} else {
			R(P) += 2;
		}
		break;
	case 0xcc: // LSlE CC	IF IE = 1, R(P) + 2 -> R(P), ELSE CONTINUE
		if (IE == 1)
			R(P) += 2;
		break;
	case 0xcd: // LSQ CD	IF Q = 1, R(P) + 2 -> R(P), ELSE CONTINUE
		if (Q == 1)
			R(P) += 2;
		break;
	case 0xce: // LSZ CE	IF D = 0, R(P) + 2 -> R(P), ELSE CONTINUE
		if (D == 0)
			R(P) += 2;
		break;
	case 0xcf: // LSDF CF	IF DF = 1, R(P) + 2 -> R(P), ELSE CONTINUE
		if (DF == 1)
			R(P) += 2;
		break;

	case 0xd0: // SEP DN	N -> P
	case 0xd1: 
	case 0xd2:
	case 0xd3:
	case 0xd4:
	case 0xd5:
	case 0xd6:
	case 0xd8:
	case 0xd9:
	case 0xda:
	case 0xdb:
	case 0xdc:
	case 0xdd:
	case 0xde:
	case 0xdf:
	case 0xd7: 
		P = N;
		break;

	case 0xe0: // SEX EN	N -> X
	case 0xe1: 
	case 0xe2:
	case 0xe3:
	case 0xe4:
	case 0xe5:
	case 0xe6:
	case 0xe7:
	case 0xe8:
	case 0xe9:
	case 0xea:
	case 0xeb:
	case 0xec:
	case 0xed:
	case 0xee:
	case 0xef:
		X = N;
		break;

	case 0xf0: // LDX		M(R(X)) -> D
		D = M(R(X));
		break;
	case 0xf1: // OR F1		M(R(X)) OR D -> D
		D |= M(R(X));
		break;
	case 0xf2: // AND F2	M(R(X)) AND D -> D
		D &= M(R(X));
		break;
	case 0xf3: // XOR F3	M(R(X)) XOR D -> D
		D ^= M(R(X));
		break;
	case 0xf4: // ADD F4	M(R(X)) + D -> DF, D
		TempLong = D + M(R(X));
		D = (unsigned char) TempLong;
		DF = TempLong >= 0x100 ? 1 : 0;
		break;
	case 0xf5: // SD F5		M(R(X)) - D -> DF, D
		DF = M(R(X)) < D ? 0 : 1;
		D = M(R(X)) - D;
		break;
	case 0xf6: // SHR F6	SHIFT D RIGHT, LSB(D) -> DF, 0 -> MSB(D)
		DF = D & 1;
		D >>= 1;
		break;
	case 0xf7: // SM F7		D-M(R(X)) -> DF, D
		DF = M(R(X)) > D ? 0 : 1;
		D -= M(R(X));
		break;

	case 0xf8: // LDl F8	M(R(P)) -> D; R(P) + 1 -> R(P)
		D = M(R(P));
		R(P)++;
		break;
	case 0xf9: // ORl F9	M(R(P)) OR D -> D; R(P) + 1 -> R(P)
		D |= M(R(P));
		R(P)++;
		break;
	case 0xfa: // ANl FA	M(R(P)) AND D -> D; R(P) + 1 -> R(P)
		D &= M(R(P));
		R(P)++;
		break;
	case 0xfb: // XRI FB	M(R(P)) XOR D -> D; R(P) + 1 -> R(P)
		D ^= M(R(P));
		R(P)++;
		break;
	case 0xfc: // ADI FC	M(R(P)) + D -> DF, D; R(P) + 1 -> R(P)
		TempLong = D + M(R(P));
		D = (unsigned char) TempLong;
		DF = TempLong >= 0x100 ? 1 : 0;
		R(P)++;
		break;
	case 0xfd: // SDl FD	M(R(P)) - D -> DF, D; R(P) + 1 -> R(P)
		DF = M(R(P)) < D ? 0 : 1;
		D = M(R(P)) - D;
		R(P)++;
		break;
	case 0xfe: // SHL FE	SHIFT D LEFT, MSB(D) -> DF, 0 -> LSB(D)
		DF = D & 0x80 ? 1 : 0;
		D <<= 1;
		break;
	case 0xff: // SMI FF	D-M(R(P)) -> DF, D; R(P) + 1 -> R(P)
		DF = M(R(P)) > D ? 0 : 1;
		D -= M(R(P));
		R(P)++;
		break;

#if DEBUG
	default:
		RVal = 1;
		break;
#endif

	}

#ifdef DEBUG
	if (DF > 1 || R(P) >= 0xf000 || X > 0xf || P > 0xf)
		RVal = 1;	// Checked
#endif

	return RVal;
}


