//
// Copyright 2000, Kirk Davis (kbd@ndx.net)
//
// Use and distribute freely. Share your changes for others.
// Use at your own risk - no warrantee expressed or implied
// (but you knew that).
//

#include "stdio.h"
#include "string.h"
#include "ctype.h"

#include "1802.h"
#include "fileio.h"

void dumpBuffer(unsigned short offset);
void DumpReg();
void help();

struct fileHeader fileHeader;

main(int ac, char **av) {
	FILE *fp;
	int argCount;
	unsigned char c;
	unsigned short i;
	unsigned long arg1;
	unsigned long arg2;
	unsigned char buf[80];
	unsigned char disBuf[80];
	unsigned char fileName[40];
	unsigned long steps ;
	static unsigned short offset = 0;

	printf("\nCOSMAC 1802 Simulator\n\n");

	while (1) {
		putchar('-');
		gets(buf);

		argCount = sscanf(buf, "%c %x %x", &c, &arg1, &arg2);

		if (argCount < 0)
			argCount = 1;

		switch(c) {

		case 'h':
			help();
			break;

		case 'r':
			if (argCount > 2) {
				R(arg1 & 0xf) = (unsigned short) arg2;
			} else
				DumpReg();
			break;

		case 'S':
			if (sscanf(buf, "%c %s %x %x", &c, &fileName, &arg1, &arg2) != 4) {
				printf("Usage: S [filename] [location] [length]\n");
				break;
			}

			if (!strchr(fileName, '.'))
				strcat(fileName, ".elf");

			if (!(fp = fopen(fileName, "wb"))) {
				perror("Open");
				break;
			}


			fileHeader.signature = SIGNATURE;
			fileHeader.version = VERSION;
			fileHeader.start = (unsigned short) arg1;
			fileHeader.length = (unsigned short) arg2;

			if (fwrite(&fileHeader, sizeof(fileHeader), 1, fp) != 1) {
				perror("write");
				break;
			}

			if (fwrite(&M(fileHeader.start), fileHeader.length, 1, fp) != 1) {
				perror("write");
				break;
			}

			fclose(fp);
			
			break;

		case 'L':
			if (sscanf(buf, "%c %s", &c, &fileName) != 2) {
				printf("Usage: L [filename]\n");
				break;
			}

			if (!strchr(fileName, '.'))
				strcat(fileName, ".elf");

			if (!(fp = fopen(fileName, "rb"))) {
				perror("Open");
				break;
			}

			if (fread(&fileHeader, sizeof(fileHeader), 1, fp) != 1) {
				perror("Error reading file header");
				break;
			}

			if (fileHeader.signature != SIGNATURE
				|| fileHeader.version != VERSION) {
				printf("Invalid file\n");
				break;
			}

			if ((arg2 = fread(&M(fileHeader.start), fileHeader.length, 1, fp)) != 1) {
				perror("Error reading file");
				break;
			}

			fclose(fp);
			
			break;

		case 'm':
			if (argCount > 2) {
				offset = (unsigned short) arg1;
				M(offset++) = (unsigned char) arg2;
			} else if (argCount > 1) {
				M(offset++) = (unsigned char) arg1;
			} else
				printf("usage: m [address] [value]\n");
			break;

		case 's':
				if (argCount > 1)
					R(P) = offset;
				
				Step();
				DumpReg();
				DisAsm(disBuf, R(P));
				printf("%s", disBuf);
				offset = R(P);
			break;

		case 'u':
			if (argCount > 1)
				offset = (unsigned short) arg1;

			for (i = 0; i < 0x10; i++ ) {
				offset += DisAsm(disBuf, offset);
				printf("%s", disBuf);
			}
			break;


		case 'd':
			if (argCount > 1)
				offset = (unsigned short) arg1;

			dumpBuffer(offset);
			offset += 0x100;
			break;

		case 'q':
			return(0);
			break;

		case 'g':
			steps = 0;

			if (argCount > 1)
				R(P) = (unsigned short) arg1;

			while (M(R(P)) != 0) {
				steps++;
				if (Step() || (buf[0] == 'z' && steps > 370)) {
					printf("Break\n\n");
					break;
				}
			}

			printf("%d instructions executed\n", steps);

			DumpReg();
			DisAsm(disBuf, R(P));
			printf("%s", disBuf);
			break;

		default:
			help();
			break;
		}
	}
}

void help() {
	printf("\nCommands:\n\n");
	printf(" h				This message\n");
	printf(" m [location] [value]		Change memory\n");
	printf(" r [reg] [value]		Display or change registers\n");
	printf(" s [location]			Single Step\n");
	printf(" g [location]			Go\n");
	printf(" u [location]			Disassemble\n");
	printf(" d [location]			Dump memory\n");
	printf(" S [file] [location] [length]	Save Memory to a file\n");
	printf(" L [file]			Load Memory from a file\n");
	printf(" q				Quit\n");
	printf("\n\n");
}

void DumpReg() {
	printf("\n");
	printf(" R0 = %0.4x, R4 = %0.4x, R8 = %0.4x, RC = %0.4x\n", R(0), R(4), R(8), R(0xc));
	printf(" R1 = %0.4x, R5 = %0.4x, R9 = %0.4x, RD = %0.4x\n", R(1), R(5), R(9), R(0xd));
	printf(" R2 = %0.4x, R6 = %0.4x, RA = %0.4x, RE = %0.4x\n", R(2), R(6), R(0xa), R(0xe));
	printf(" R3 = %0.4x, R7 = %0.4x, RB = %0.4x, RF = %0.4x\n", R(3), R(7), R(0xb), R(0xf));
	printf("\n");
	printf(" D = %0.2x, DF = %0.1x, B = %0.2x, P = %0.1x, X = %0.1x\n", D, DF, B, P, X);
	printf(" Q = %0.1x, N = %0.1x, I = %0.1x", Q, N, I);
	printf("\n\n");
}

void dumpBuffer(unsigned short offset) {
	int i;
	int ii;

	for (i = 0; i < 16; i++) {
		printf("%0.4x: ", offset + (i * 0x10));
		for (ii = 0; ii < 0x10; ii++) {
			printf("%0.2x", M(offset + (i * 0x10) + ii));
			if (M(offset + (i * 0x10) + ii) == 0xff)
				printf(":");
			else
				printf(" ");
		}

		printf("  |");
		for (ii = 0; ii < 0x10; ii++) 
			printf("%c", isprint(M(offset + (i * 0x10) + ii)) ?
			M(offset + (i * 0x10) + ii) : '.');

		printf("|\n");
	}
}
